﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Web.UI;
using HtmlAgilityPack;
using Microsoft.SharePoint.Utilities;

namespace RH.WikiNavigation.DataClasses
{
    public class WikiData : IHierarchyData
    {
        public int Id { get; set; }

        /// <summary>
        /// e.g. New%20Page.aspx
        /// </summary>
        public string FileName { get; set; }

        /// <summary>
        /// The Html which is stored in the WikiField
        /// </summary>
        public string WikiField { get; set; }

        /// <summary>
        /// e.g. /wiki/wiki%20pages/New%20Page.aspx
        /// </summary>
        public IList<string> Links { get; set; }

        /// <summary>
        /// wiki pages, which are linked in the WikiField
        /// </summary>
        public IList<WikiData> Children { get; set; }

        public WikiData Parent { get; set; }

        private WikiData Root { get; set; }

        private readonly string _ListUrl;
        private readonly WikiNavigationAdapter _Adapter;

        #region Properties for DataBinding
        public string Text { get
        {
            if (string.IsNullOrEmpty(FileName))
            {
                _Adapter.WriteULSVerbose("Text: The FileName is empty");
                return string.Empty;
            }
            return FileName.Replace(".aspx", string.Empty);
        } }
        public string Value { get { return Id.ToString(); } }
        public string ImageUrl { get { return string.Empty;}}
        public string NavigateUrl { get { return _ListUrl + "/" + FileName; } }
        #endregion

        public WikiData(WikiData parent, DataRow row, string listUrl, WikiNavigationAdapter adapter)
        {
            _Adapter = adapter;
            Links = new List<string>();
            Children = new List<WikiData>();
            _ListUrl = listUrl;
            if (row == null)
            {
                _Adapter.WriteULSVerbose("row is null");
                return;
            }

            Id = (int) row["ID"];
            _Adapter.WriteULSVerbose("ID" + Id);
            FileName = (row["FileLeafRef"] != DBNull.Value) ? (string) row["FileLeafRef"] : string.Empty;
            _Adapter.WriteULSVerbose("FileName" + FileName);
            WikiField = (row["WikiField"] != DBNull.Value) ? (string) row["WikiField"] : string.Empty;
            _Adapter.WriteULSVerbose("WikiField" + WikiField);
            Parent = parent;

            // find root
            Root = this;
            WikiData temp = Parent;
            while (temp != null)
            {
                Root = temp;
                temp = temp.Parent;
            }
            _Adapter.WriteULSVerbose("Root WikiData" + Root.FileName);

            ExtractLinks(_ListUrl);
        }

        /// <summary>
        /// find links to other wiki pages within the WikiField
        /// </summary>
        /// <param name="listUrl"></param>
        private void ExtractLinks(string listUrl)
        {
            try
            {
                if (string.IsNullOrEmpty(WikiField))
                {
                    _Adapter.WriteULSVerbose("WikiField is empty");
                    return;
                }

                var oHtml = new HtmlDocument();
                _Adapter.WriteULSVerbose("Loading WikiField");
                oHtml.LoadHtml(WikiField);
                HtmlNodeCollection oCollection = oHtml.DocumentNode.SelectNodes("//a");

                if (oCollection == null)
                {
                    _Adapter.WriteULSVerbose("No link found");
                    return;
                }

                foreach (HtmlNode oNode in oCollection)
                {
                    var href = oNode.Attributes["href"].Value;
                    _Adapter.WriteULSVerbose("parsing link: " + href);
                    var decodedHref = SPEncode.UrlDecodeAsUrl(href);
                    _Adapter.WriteULSVerbose("decodedHref" + decodedHref);
                    var decodedListUrl = SPEncode.UrlDecodeAsUrl(listUrl);
                    _Adapter.WriteULSVerbose("decodedListUrl" + decodedListUrl);
                    if (!decodedHref.StartsWith(decodedListUrl) || Root.ContainsLink(href)) continue;

                    _Adapter.WriteULSVerbose("Adding link to Links");
                    Links.Add(href);
                }
            }
            catch (Exception ex)
            {
                _Adapter.WriteULSError(string.Format("Error Extracting Links. WikiField: {0} ex: {1}", WikiField, ex));
            }
        }

        #region Implementation of IHierarchyData
        public IHierarchicalEnumerable GetChildren()
        {
            // Call to the local cache for the data
            var children = new WikiDataDataSource();

            // Loop through your local data and find any children
            foreach (WikiData child in Children)
            {
                children.Add(child);
            }

            return children;
        }

        public IHierarchyData GetParent()
        {
            return Parent;
        }

        public bool HasChildren
        {
            get { return Children.Count > 0; }
        }

        public string Path
        {
            get { return Id.ToString(); }
        }

        public object Item
        {
            get { return this; }
        }

        public string Type
        {
            get { return GetType().ToString(); }
        }
        #endregion
    }
}