﻿using System;
using System.IO;
using System.Reflection;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using Microsoft.SharePoint.Utilities;

namespace RH.WikiNavigation
{
    /// <summary>
    /// with code from http://community.zevenseas.com/Blogs/Robin/archive/2009/01/21/feature-blocker.aspx
    /// </summary>
    public class FeatureReceiver : SPFeatureReceiver
    {
        //Assembly
        private const string BinPath = "bin";
        //Browser file
        private const string AppBrowserPath = "App_Browsers";
        private const string BrowserFile = "RHWikiNavigation.browser";
        private const string FeatureName = "RHWikiNavigation";

        public override void FeatureActivated(SPFeatureReceiverProperties properties)
        {
            InstallBrowserFile(properties, InstallMode.Install);
        }

        public override void FeatureDeactivating(SPFeatureReceiverProperties properties)
        {
            InstallBrowserFile(properties, InstallMode.Uninstall);
        }

        public override void FeatureInstalled(SPFeatureReceiverProperties properties)
        {
        }

        public override void FeatureUninstalling(SPFeatureReceiverProperties properties)
        {
            try
            {
                var currentWebApplication = (SPWebApplication)properties.Feature.Parent;
                SPFarm farm = currentWebApplication.Farm;
                var service = farm.Services.GetValue<SPWebService>("");

                foreach (SPWebApplication webApplication in service.WebApplications)
                {
                    InstallBrowserFile(webApplication, InstallMode.Uninstall);
                }
            }
            catch (Exception ex)
            {
                throw new SPException("Error uninstalling the feature from the farm: " + ex);
            }

        }

        /// <summary>
        /// store/remove browser file in App_Browsers Directory
        /// </summary>
        /// <param name="webApplication"></param>
        /// <param name="installMode"></param>
        private void InstallBrowserFile(SPWebApplication webApplication, InstallMode installMode)
        {
            try
            {
                TraceProvider.RegisterTraceProvider();
                if (webApplication == null) return;
                foreach (SPUrlZone urlZone in Enum.GetValues(typeof(SPUrlZone)))
                {
                    WriteULSVerbose("Working on UrlZone " + Enum.GetName(typeof(SPUrlZone), urlZone));
                    SPIisSettings iisSettings = webApplication.GetIisSettingsWithFallback(urlZone);
                    ModifyZone(iisSettings, installMode);
                }
            }
            catch (Exception ex)
            {
                WriteULSVerbose("Error: "+ex);
                throw new SPException("Can not install browser file: " + ex);
            }
            finally
            {
                TraceProvider.UnregisterTraceProvider();
            }
        }

        private void ModifyZone(SPIisSettings iisSettings, InstallMode installMode)
        {
            //Browser file
            string browserFileSourcePath = String.Format("{0}\\FEATURES\\{1}\\{2}", SPUtility.GetGenericSetupPath("Template"), FeatureName, BrowserFile);
            WriteULSVerbose("browserFileSourcePath: "+browserFileSourcePath);
            string browserFileDestination = String.Format("{0}\\{1}\\{2}", iisSettings.Path, AppBrowserPath, BrowserFile);
            WriteULSVerbose("browserFileDestination: "+browserFileDestination);
            AddRemoveFile(installMode, browserFileSourcePath, browserFileDestination);

            // ensure bin directoy
            DirectoryInfo[] binFolder = iisSettings.Path.GetDirectories(BinPath);
            if (binFolder.Length == 0)
            {
                WriteULSVerbose("Creating directory "+BinPath);
                iisSettings.Path.CreateSubdirectory(BinPath);
            }

            //Assembly
            Assembly assembly = GetType().Assembly;
            if (assembly.Location == null) throw new FileNotFoundException("File " + assembly.FullName + " not found!");

            var assemblyFileDestination = String.Format("{0}\\{1}\\{2}", iisSettings.Path, BinPath, Path.GetFileName(assembly.Location));
            WriteULSVerbose("assemblyFileDestination: "+assemblyFileDestination);
            if (!File.Exists(assemblyFileDestination))
            {
                AddRemoveFile(installMode, assembly.Location, assemblyFileDestination);
            }
        }

        /// <summary>
        /// store/remove browser file in App_Browsers Directory
        /// </summary>
        /// <param name="properties"></param>
        /// <param name="installMode"></param>
        private void InstallBrowserFile(SPFeatureReceiverProperties properties, InstallMode installMode)
        {
            var webApplication = properties.Feature.Parent as SPWebApplication;
            InstallBrowserFile(webApplication, installMode);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="installMode"></param>
        /// <param name="sourcePath">null if installMode=Uninstall</param>
        /// <param name="destinationPath"></param>
        private static void AddRemoveFile(InstallMode installMode, string sourcePath, string destinationPath)
        {
            if (installMode == InstallMode.Install)
            {
                // add browser file
                var fileInfo = new FileInfo(sourcePath);
                if (fileInfo.Exists == false) throw new FileNotFoundException("File " + sourcePath + " not found!");
                
                WriteULSVerbose("Install: adding file "+destinationPath);
                fileInfo.CopyTo(destinationPath, true);
            }
            else
            {
                // remove browser file
                var fileInfo = new FileInfo(destinationPath);
                if (fileInfo.Exists)
                {
                    WriteULSVerbose("Uninstall: deleting file " + destinationPath);
                    fileInfo.Delete();
                }
            }
        }

        private static void WriteULSVerbose(string message)
        {
            try
            {
                TraceProvider.WriteTrace(0, TraceProvider.TraceSeverity.Verbose, Guid.Empty, "RH.WikiNavigation.dll", "RH.WikiNavigation", "General", message);
            }
// ReSharper disable EmptyGeneralCatchClause
            catch
// ReSharper restore EmptyGeneralCatchClause
            {
                // ignore any errors regarding writing to the ULS logs
            }
        }
    }
}
