﻿using System;
using System.Collections.Specialized;
using System.Data;
using System.Linq;
using System.Web.Caching;
using System.Web.UI;
using System.Web.UI.Adapters;
using System.Web.UI.HtmlControls;
using Microsoft.SharePoint;
using System.Web.UI.WebControls;
using Microsoft.SharePoint.Utilities;
using Microsoft.SharePoint.WebControls;
using RH.WikiNavigation.DataClasses;
using System.Web;

namespace RH.WikiNavigation
{
    /// <summary>
    /// QueryString Refresh=true will force a reload and not use cached data
    /// QueryString ExpandDepth=x will change the expand depth of the treeview
    /// QueryString MaxLength=x will shorten the TreeNode text
    /// QueryString Debug=true will write information to the SharePoint Logs
    /// </summary>
    // ReSharper disable UnusedMember.Global
    public class WikiNavigationAdapter : ControlAdapter
    // ReSharper restore UnusedMember.Global
    {
        private HyperLink _HeaderLink;
        private SPTreeView _Navigation;
        private Control _PlaceHolderLeftNavBar;
        private static readonly object Lock = new object();
        private readonly Cache _Cache = HttpRuntime.Cache;
        private DataTable _DataTable;
        private WikiData _NavigationData;
        private string _ListUrl;
        private string _HomepageUrl;
        private bool _Debug;

        protected override void CreateChildControls()
        {
            base.CreateChildControls();

            try
            {
                string debug = Page.GetQueryString("Debug");
                if (!string.IsNullOrEmpty(debug) && Boolean.TryParse(debug, out _Debug))
                {
                    TraceProvider.RegisterTraceProvider();
                }

                SPWeb web = SPContext.Current.Web;
                SPList list = SPContext.Current.List;
                _ListUrl = string.Format("{0}/{1}", list.ParentWebUrl, SPEncode.HtmlEncode(list.RootFolder.Url));
                WriteULSVerbose("ListUrl: " + _ListUrl);
                _PlaceHolderLeftNavBar = Page.Form.FindControl("PlaceHolderLeftNavBar");
                if (_PlaceHolderLeftNavBar == null)
                {

                    Page.Form.Controls.Add(new Literal { Text = "Can not find control with ID PlaceHolderLeftNavBar" });
                    return;
                }

                FindHomepageUrl(web, list);

                CreateNavigationControls(web);
            }
            catch (Exception ex)
            {
                ShowError("Error in WikiNavigationAdapter: " + ex);
            }
        }

        protected override void OnPreRender(EventArgs e)
        {
            base.OnPreRender(e);
            try
            {
                SPWeb web = SPContext.Current.Web;
                SPList list = SPContext.Current.List;

                CacheData(web, list);
                FillTreeView();

                if (_Debug)
                {
                    TraceProvider.UnregisterTraceProvider();
                }
            }
            catch (Exception ex)
            {
                ShowError("Error in OnPreRender: " + ex);
            }
        }

        /// <summary>
        /// find the homepage url from the web. if it does not exist, use the first item in the wiki library
        /// the homepage url will be library/page.aspx
        /// </summary>
        /// <param name="web"></param>
        /// <param name="currentList"></param>
        private void FindHomepageUrl(SPWeb web, SPList currentList)
        {
            // homepage e.g. /wiki pages/homepage.aspx
            _HomepageUrl = web.RootFolder.WelcomePage;
            WriteULSVerbose("Homepage: " + _HomepageUrl);
            if (_HomepageUrl.IndexOf('/') == -1)
            {
                // did not find a homepage. maybe we are inside a wiki library, but not a wiki page template
                bool wikiEnabled = currentList.ContentTypes.Cast<SPContentType>().Any(c => c.Parent.Id == SPBuiltInContentTypeId.WikiDocument);

                if (wikiEnabled)
                {
                    WriteULSVerbose("We are inside a wiki library");
                    // use the first item as homepage
                    var query = new SPQuery { RowLimit = 1 };
                    var items = currentList.GetItems(query);
                    if (items.Count == 1)
                    {
                        _HomepageUrl = string.Format("{0}/{1}", currentList.RootFolder.Url, items[0][SPBuiltInFieldId.FileLeafRef]);
                        WriteULSVerbose("Using " + _HomepageUrl + " as homepage");
                    }
                }
                else
                {
                    WriteULSError("No homepage found");
                    return;
                }
            }
        }

        /// <summary>
        /// fill TreeView
        /// </summary>
        private void FillTreeView()
        {
            if (_NavigationData == null)
            {
                WriteULSError("Homepage not found!");
                _Navigation.Nodes.Add(new TreeNode("Homepage not found!"));
                return;
            }

            if (_Navigation.Nodes.Count != 0)
            {
                WriteULSVerbose("no navigation nodes found");
                return;
            }
            string currentItemId = SPContext.Current.ListItem.ID.ToString();
            WriteULSVerbose("CurrentItemId: " + currentItemId);
            int? maxLength = GetMaxLength();
            WriteULSVerbose("maxLength: " + maxLength);
            int expandDepth = GetExpandDepth();
            WriteULSVerbose("expandDepth" + expandDepth);

            // do not add the rootnode
            foreach (WikiData childNode in _NavigationData.Children)
            {
                WriteULSVerbose("Adding node: " + childNode.NavigateUrl);
                // expand this level
                var node = new TreeNode(GetNodeText(childNode.Text, maxLength), childNode.Value, childNode.ImageUrl)
                {
                    NavigateUrl = AddQueryString(childNode.NavigateUrl, Page.Request),
                    Expanded = expandDepth >= 1
                };
                _Navigation.Nodes.Add(node);

                if (childNode.Value == currentItemId)
                {
                    node.Selected = true;
                    WriteULSVerbose("Expanding node and parent");
                    ExpandNodesUntilSelected(node);
                }
                AddNodes(node, childNode, currentItemId, Page.Request, maxLength, expandDepth, 1);
            }
        }

        /// <summary>
        /// add a treenode for each wikidata
        /// </summary>
        /// <param name="node"></param>
        /// <param name="data"></param>
        /// <param name="currentItemId"></param>
        /// <param name="request"></param>
        /// <param name="maxLength"></param>
        /// <param name="expandDepth"></param>
        /// <param name="depth"></param>
        private void AddNodes(TreeNode node, WikiData data, string currentItemId, HttpRequest request, int? maxLength, int expandDepth, int depth)
        {
            depth++;
            WriteULSVerbose("depth: " + depth);
            // the node has been added before. so only add the childnodes
            foreach (WikiData childData in data.Children)
            {
                WriteULSVerbose("Adding node: " + childData.NavigateUrl);
                var childNode = new TreeNode(GetNodeText(childData.Text, maxLength), childData.Value, childData.ImageUrl)
                {
                    NavigateUrl = AddQueryString(childData.NavigateUrl, request),
                    Expanded = expandDepth >= depth
                };
                node.ChildNodes.Add(childNode);

                if (childData.Value == currentItemId)
                {
                    childNode.Selected = true;
                    WriteULSVerbose("Expanding node and parent");
                    ExpandNodesUntilSelected(childNode);
                }

                AddNodes(childNode, childData, currentItemId, request, maxLength, expandDepth, depth);
            }
        }

        /// <summary>
        /// shorten node text if maxlength is not null and the nodetext is longer than maxlength
        /// </summary>
        /// <param name="nodeText"></param>
        /// <param name="maxLength"></param>
        /// <returns></returns>
        private static string GetNodeText(string nodeText, int? maxLength)
        {
            if (maxLength != null && nodeText.Length > maxLength)
            {
                return nodeText.Substring(0, (int)maxLength) + "...";
            }
            return nodeText;
        }

        /// <summary>
        /// expand parent nodes for the selected node, so it is always visible
        /// </summary>
        /// <param name="node"></param>
        private static void ExpandNodesUntilSelected(TreeNode node)
        {
            node.Expand();
            while (node.Parent != null)
            {
                node = node.Parent;
                node.Expand();
            }
        }

        /// <summary>
        /// get the home wiki page from the cached _DataTable
        /// </summary>
        /// <returns>WikiData | null</returns>
        private WikiData FindRootWikiPage()
        {
            // we only need the filename here
            _HomepageUrl = _HomepageUrl.Substring(_HomepageUrl.LastIndexOf('/') + 1);

            WriteULSVerbose("Homepage (file only): " + _HomepageUrl);
            DataRow homeRow = _DataTable.Rows.Cast<DataRow>().SingleOrDefault(r => r["FileLeafRef"] != DBNull.Value && Convert.ToString(r["FileLeafRef"]) == _HomepageUrl);
            if (homeRow == null)
            {
                WriteULSVerbose("No homeRow found");
                return null;
            }

            var home = new WikiData(null, homeRow, _ListUrl, this);
            WriteULSVerbose("Got home: " + home.FileName);
            return home;
        }

        /// <summary>
        /// recursively find all Wiki pages, which are linked in the current WikiField and add them to the children IList
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        private void AddLinkedWikiPages(WikiData item)
        {
            if (item == null) throw new ArgumentException("The WikiData parameter must not be null!", "item");
            if (item.Links == null || _DataTable == null || _DataTable.Rows == null || !_DataTable.Columns.Contains("FileLeafRef")) return;
            try
            {
                foreach (string link in item.Links)
                {
                    if (link == null) continue;
                    // FileLeafRef e.g. "New Page.aspx"
                    // link e.g. /wiki/wiki%20pages/New%20Page.aspx
                    int marker = link.LastIndexOf('/');
                    string wikiPageName = (marker != -1) ? link.Substring(marker + 1) : link;
                    wikiPageName = SPEncode.UrlDecodeAsUrl(wikiPageName);
                    DataRow row = _DataTable.Rows.Cast<DataRow>().SingleOrDefault(r => r["FileLeafRef"] != DBNull.Value && Convert.ToString(r["FileLeafRef"]) == wikiPageName);

                    if (row == null) continue;
                    var data = new WikiData(item, row, _ListUrl, this);
                    item.Children.Add(data);

                    // now recusively add all wiki pages
                    AddLinkedWikiPages(data);
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Error in AddLinkedWikiPages for item with ID " + item.Id, ex);
            }
        }

        /// <summary>
        /// read pages from wiki and store them in _DataTable
        /// </summary>
        /// <param name="web"></param>
        /// <param name="list"></param>
        private void CacheData(SPWeb web, SPList list)
        {
            string cacheTime = string.Format("CacheTime{0}", list.ID);
            string wikiDataCacheName = string.Format("WikiDataWebID{0}", list.ID);

            lock (Lock)
            {
                string queryString = Page.GetQueryString("Refresh");
                bool needRefresh = queryString == "TRUE" || ListItemsChanged(list, _Cache[cacheTime]);
                _NavigationData = _Cache[wikiDataCacheName] as WikiData;
                if (_NavigationData != null && !needRefresh)
                {
                    WriteULSVerbose("No cached data (Refresh=" + queryString + ")");
                    return;
                }

                SPSecurity.RunWithElevatedPrivileges(() => GetListItems(web, list, wikiDataCacheName, cacheTime));
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="web">elevated SPWeb</param>
        /// <param name="list">wiki list</param>
        /// <param name="wikiDataCacheName"></param>
        /// <param name="cacheTime"></param>
        private void GetListItems(SPWeb web, SPList list, string wikiDataCacheName, string cacheTime)
        {
            // clear cache
            _Cache.Remove(wikiDataCacheName);
            _Cache.Remove(cacheTime);

            // now get items and store them again
            using (var elevatedSite = new SPSite(web.Url))
            using (var elevatedWeb = elevatedSite.OpenWeb())
            {
                var elevatedList = elevatedWeb.Lists[list.ID];
                var query = new SPQuery { Query = "<Where><IsNull><FieldRef Name='CheckoutUser' /></IsNull></Where>" };

                // read items from list
                WriteULSVerbose("Fetching items");
                _DataTable = elevatedList.GetItems(query).GetDataTable();

                // work on data
                _NavigationData = FindRootWikiPage();
                if (_NavigationData != null)
                {
                    AddLinkedWikiPages(_NavigationData);
                    WriteULSVerbose("Adding NavigationData to cache");
                    _Cache.Add(wikiDataCacheName,
                               _NavigationData,
                               null,
                               DateTime.Now.AddMinutes(20),
                               Cache.NoSlidingExpiration,
                               CacheItemPriority.Normal,
                               null);
                    _Cache[cacheTime] = DateTime.Now;
                }
            }
        }

        /// <summary>
        /// check if the data has changed, since it was cached
        /// </summary>
        /// <param name="list"></param>
        /// <param name="cacheDateTime"></param>
        /// <returns></returns>
        private static bool ListItemsChanged(SPList list, object cacheDateTime)
        {
            if (cacheDateTime == null) return true;
            var token = new SPChangeToken(SPChangeCollection.CollectionScope.List, list.ID, ((DateTime)cacheDateTime).ToUniversalTime());
            var changeQuery = new SPChangeQuery(true, true)
            {
                ChangeTokenStart = token,
                Add = true,
                Delete = true,
                Item = true,
                Move = true,
                Rename = true,
                Restore = true,
                Update = true
            };
            SPChangeCollection collection = list.GetChanges(changeQuery);
            return collection.Count > 0;
        }

        private void CreateNavigationControls(SPWeb web)
        {
            int expandDepth = 2;
            _Navigation = new SPTreeView { ID = "WikiNavigationTreeView", ExpandDepth = expandDepth, CssClass = "ms-vb", NodeIndent = 3 };
            if (int.TryParse(Page.GetQueryString("ExpandDepth"), out expandDepth)) _Navigation.ExpandDepth = expandDepth;
            WriteULSVerbose("ExpandDepth: " + _Navigation.ExpandDepth);
            _Navigation.SelectedNodeStyle.CssClass = "ms-tvselected";
            _Navigation.HoverNodeStyle.CssClass = "ms-tvselected";

            var outer = new HtmlGenericControl("DIV");
            outer.Attributes.Add("class", "ms-quicklaunchouter");

            var quicklaunch = new HtmlGenericControl("DIV");
            quicklaunch.Attributes.Add("class", "ms-navsubmenu1");
            quicklaunch.Style.Add("style", "width:100%");
            outer.Controls.Add(quicklaunch);

            var quicklauunchheader = new HtmlGenericControl("DIV");
            quicklauunchheader.Attributes.Add("class", "ms-quicklaunchheader");
            quicklaunch.Controls.Add(quicklauunchheader);

            string homepageUrl = string.Format("{0}/{1}", web.ServerRelativeUrl, _HomepageUrl);
            _HeaderLink = new HyperLink { NavigateUrl = AddQueryString(homepageUrl, Page.Request), Text = "Wiki Navigation", ID = "WikiNavigationHeader" };
            quicklauunchheader.Controls.Add(_HeaderLink);

            // navigation belongs here
            var tableDiv = new HtmlGenericControl("DIV");
            tableDiv.Style.Add("overflow", "auto");
            tableDiv.Style.Add("height", "400px");
            tableDiv.Style.Add("width", "150px;");
            var table = new Table { CssClass = "ms-navSubMenu1", CellPadding = 0, CellSpacing = 0, BorderStyle = BorderStyle.None };
            var row = new TableRow();
            var cell = new TableCell();
            cell.Controls.Add(_Navigation);
            row.Cells.Add(cell);
            table.Rows.Add(row);
            tableDiv.Controls.Add(table);
            quicklaunch.Controls.Add(tableDiv);

            // add custom controls to control collection of page
            _PlaceHolderLeftNavBar.Controls.Add(outer);
        }

        /// <summary>
        /// get the maximum node text length from the querystring
        /// </summary>
        /// <returns></returns>
        private int? GetMaxLength()
        {
            string queryString = Page.GetQueryString("MaxLength");
            int maxLength;
            if (queryString != null && int.TryParse(queryString, out maxLength))
            {
                return maxLength;
            }
            return null;
        }

        /// <summary>
        /// get expand depth from querystring. if not specified, default is 2
        /// </summary>
        /// <returns></returns>
        private int GetExpandDepth()
        {
            string queryString = Page.GetQueryString("ExpandDepth");
            int maxLength;
            if (queryString != null && int.TryParse(queryString, out maxLength))
            {
                return maxLength;
            }
            return 2;
        }

        private void ShowError(string errorText)
        {
            Control control = _Navigation.Parent;
            _Navigation.Visible = false;
            control.Controls.Add(new LiteralControl(errorText));
            WriteULSError(errorText);
        }

        internal static string AddQueryString(string url, HttpRequest request)
        {
            if (request.QueryString == null || !request.QueryString.HasKeys()) return url;
            NameValueCollection queryStrings = request.QueryString;

            // remove existing querystrings
            int marker = url.IndexOf('?');
            if (marker > 0) url = url.Substring(0, marker);

            // add querystring
            if (!queryStrings.HasKeys()) return url;

            for (int i = 0; i < queryStrings.Keys.Count; i++)
            {
                string key = queryStrings.AllKeys[i];
                string value = queryStrings[i];
                // don't add Refresh
                if (string.Compare(key, "refresh", StringComparison.OrdinalIgnoreCase) == 0) continue;

                url += url.Contains("?") ? "&" : "?";
                url += string.Format("{0}={1}", key, value);
            }
            return url;
        }

        /// <summary>
        /// write to the ULS Log (TraceSeverity = Exception)
        /// </summary>
        /// <param name="text"></param>
        internal void WriteULSError(string text)
        {
            try
            {
                if (_Debug)
                {
                    TraceProvider.WriteTrace(0, TraceProvider.TraceSeverity.Exception, Guid.Empty, "RH.WikiNavigation.dll", "RH.WikiNavigation", "General", text);
                }
            }
// ReSharper disable EmptyGeneralCatchClause
            catch
// ReSharper restore EmptyGeneralCatchClause
            {
                // ignore any errors regarding writing to the ULS logs
            }
        }

        /// <summary>
        /// write to the ULS Log (TraceSeverity = Verbose)
        /// </summary>
        /// <param name="text"></param>
        internal void WriteULSVerbose(string text)
        {
            try
            {
                if (_Debug)
                {
                    TraceProvider.WriteTrace(0, TraceProvider.TraceSeverity.Verbose, Guid.Empty, "RH.WikiNavigation.dll", "RH.WikiNavigation", "General", text);
                }
            }
// ReSharper disable EmptyGeneralCatchClause
            catch
// ReSharper restore EmptyGeneralCatchClause
            {
                // ignore any errors regarding writing to the ULS logs
            }
        }
    }
}